#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>

#ifdef MSC
#  include <sys/types.h>
#endif

#include <sys/stat.h>
#include <time.h>
#include <ctype.h>
#include <dos.h>

#ifndef	MSC
#  include <dir.h>
#else
#  include <direct.h>
#endif

#ifdef MSC
#  define FA_DIREC  _A_SUBDIR
#  define FA_NORMAL _A_NORMAL
#  define FA_RDONLY _A_RDONLY
#  define FA_HIDDEN _A_HIDDEN
#  define FA_SYSTEM _A_SYSTEM
#  define FA_LABEL  _A_VOLID
#  define FA_ARCH   _A_ARCH
#endif

#include "../../memsl.h"

#define	SORTNAME 0
#define	SORTDATE 1
#define	SORTTIME 2
#define	SORTSIZE 3
#define	SORTACCROSS 0
#define	LISTLONG 1
#define	MAXCOL 80
#define	MAXMATCH 256
#define	MAXSTR 256

struct dir_list
  {
    char *filename;
    int	attrib,
	mode;
    long size,
	 date;
  };

struct list_info
  {
    WBSLIST *namelist;
    int	nameonly,
	charname,
	listtype,
	sortby,
	maxlen,
	col,
	reverse,
	recursive,
	diskusage;
    long totsize,
	 gtotsize,
	 totfiles,
	 gtotfiles;
  };

static char *month[] = { "Jan","Feb","Mar","Apr","May","Jun",
			 "Jul","Aug","Sep","Oct","Nov","Dec" };

int SortByName(struct list_info	*listinfo, struct dir_list *direntry1, struct dir_list *direntry2)
  {
    int	rval;

    rval = strcmp(direntry1->filename,direntry2->filename);

    if (listinfo->reverse)
      if (rval < 0)  return(1);
      else if (rval > 0) return(-1);

    return(rval);
  }

int SortByDate(struct list_info	*listinfo, struct dir_list *direntry1, struct dir_list *direntry2)
  {
    if (direntry1->date	< direntry2->date)
      return(listinfo->reverse ? 1 : -1);
    else if (direntry1->date > direntry2->date)
      return(listinfo->reverse ? -1 : 1);

    return(0);
  }

int SortByTime(struct list_info	*listinfo, struct dir_list *direntry1, struct dir_list *direntry2)
  {
    struct tm *tmtime1,	*tmtime2;
    long time1,	time2;

    tmtime1 = localtime(&direntry1->date);
    tmtime2 = localtime(&direntry2->date);
    time1 = (tmtime1->tm_hour*100)+(tmtime1->tm_min*10)+tmtime1->tm_sec;
    time2 = (tmtime2->tm_hour*100)+(tmtime2->tm_min*10)+tmtime2->tm_sec;

    if (time1 <	time2)
      return(listinfo->reverse ? 1 : -1);
    else if (time1 > time2)
      return(listinfo->reverse ? -1 : 1);

    return(0);
  }

int SortBySize(struct list_info	*listinfo, struct dir_list *direntry1, struct dir_list *direntry2)
  {
    if (direntry1->size	< direntry2->size)
      return(listinfo->reverse ? 1 : -1);
    else if (direntry1->size > direntry2->size)
      return(listinfo->reverse ? -1 : 1);

    return(0);
  }

int strmatch(char *str1, char *str2)
  {
    char tmpstr[MAXMATCH+1];
    int	found, i;

    while (*str1 && *str2)
      {
	if (*str1 == *str2)
	  {
	    str1++; str2++;
	  }
	else if	(*str1 == '?')
	  {
	    str1++;
	    if (*str1 != *str2)
	      str2++;
	  }
	else if	(*str2 == '?')
	  {
	    str2++;
	    if (*str1 != *str2)
	      str1++;
	  }
	else if	(*str1 == '*')
	  {
	    if (strcmp(str1,"*") == 0)
	      return(1);

	    str1++;

	    i =	0; found = 0;
	    while (found == 0 && i < MAXMATCH && str1[i])
	      {
		if (str1[i] == '?' || str1[i] == '*')
		  found	= 1;
		else
		  i++;
	      }

	    strncpy(tmpstr,str1,i);
	    *(tmpstr+i)	= '\0';

	    if (strstr(str2,tmpstr))
	      {
		str2 = strstr(str2,tmpstr)+i;
		str1+=i;
	      }
	    else return(0);
	  }
	else if	(*str2 == '*')
	  {
	    if (strcmp(str2,"*") == 0)
	      return(1);

	    str2++;

	    i =	0; found = 0;
	    while (found == 0 && i < MAXMATCH && str2[i])
	      {
		if (str2[i] == '?' || str2[i] == '*')
		  found	= 1;
		else
		  i++;
	      }

	    strncpy(tmpstr,str2,i);
	    *(tmpstr+i)	= '\0';

	    if (strstr(str1,tmpstr))
	      {
		str1 = strstr(str1,tmpstr)+i;
		str2+=i;
	      }
	    else return(0);
	  }
	else return(0);
      }

    if (*str1 || *str2)
      return(0);

    return(1);
  }

char *insstr(char *str1, char *str2, int pos)
  {
    int	i, i2, tmpch;

    i2 = strlen(str2);
    tmpch = str1[pos];
    for	(i = strlen(str1); i > pos; i--)
      str1[i+i2] = str1[i];
    strcpy(&str1[pos],str2);
    str1[pos+i2] = (char) tmpch;

    return(str1);
  }

char *formatlong(char *str, int	len, long num)
  {
    int	i;

    memset(str,0,len);
    ltoa(num,str,10);

    if (strlen(str) > 3)
      {
	i = strlen(str);
	while (i > 3)
	  {
	    i-=3;
	    insstr(str,",",i);
	  }
      }

    return(str);
  }

void ListDir(struct list_info *listinfo, struct	dir_list *direntry)
  {
    struct tm *tmtime;
    char charname[2], modestr[21], tmpstr[41];

    if (listinfo->charname)
      {
	if (direntry->attrib&FA_DIREC)
	  charname[0] =	'/';
	else if	(strlen(strstr(direntry->filename,".EXE")) == 4)
	  charname[0] =	'*';
	else charname[0] = ' ';
      }
    else charname[0] = '\0';
    charname[1]	= '\0';

    listinfo->totfiles++;
    listinfo->gtotfiles++;
    listinfo->totsize += direntry->size;
    listinfo->gtotsize += direntry->size;

    if (listinfo->diskusage == 0)
      switch(listinfo->listtype)
	{
	  case LISTLONG:
	    tmtime = localtime(&direntry->date);

	    memset(modestr,0,21);

	    if (direntry->attrib&FA_DIREC) modestr[0] =	'd';
	    else modestr[0] = '	';

	    if (direntry->attrib&FA_NORMAL) modestr[1] = 'n';
	    else modestr[1] = '-';

	    if (direntry->attrib&FA_RDONLY) modestr[2] = 'r';
	    else modestr[2] = '-';

	    if (direntry->attrib&FA_HIDDEN) modestr[3] = 'h';
	    else modestr[3] = '-';

	    if (direntry->attrib&FA_SYSTEM) modestr[4] = 's';
	    else modestr[4] = '-';

	    if (direntry->attrib&FA_LABEL) modestr[5] =	'l';
	    else modestr[5] = '-';

	    if (direntry->attrib&FA_ARCH) modestr[6] = 'a';
	    else modestr[6] = '-';

	    printf(" %s	  %14.14s  %s %02d,%d  %02d:%02d:%02d %s  %*s%s\n",
					   modestr,
					   formatlong(tmpstr,40,direntry->size),
					   month[tmtime->tm_mon],
					   tmtime->tm_mday,
					   1900+tmtime->tm_year,
					   tmtime->tm_hour < 12	? tmtime->tm_hour : tmtime->tm_hour-12,
					   tmtime->tm_min,
					   tmtime->tm_sec,
					   tmtime->tm_hour < 12	? "AM" : "PM",
					   listinfo->charname,
					   charname,
					   direntry->filename);
	  break;
	  case SORTACCROSS:
	    if (listinfo->col +	listinfo->maxlen + listinfo->charname +	1 >= MAXCOL)
	      {
		listinfo->col =	0;
		printf("\n");
	      }

	    listinfo->col += listinfo->maxlen +	listinfo->charname + 1;

	    printf("%*s%-*.*s",listinfo->charname,
			       charname,
			       listinfo->maxlen+1,
			       listinfo->maxlen+1,
			       direntry->filename);
	  break;
	}
  }

void FreeList(void *nullitem, struct dir_list *direntry)
  {
    nullitem = nullitem;

    free(direntry->filename);
    free(direntry);
  }

void GetDir(struct list_info *listinfo,	char *matchstr)
  {
    WBAVL *avltree;
    struct dir_list *avlentry;
    struct find_t direntry;
    struct stat	filestat;
    char dirstr[MAXSTR+1], *strptr;
    char tmpstr[41];

    if ((avltree = WBAVLOpen(NULL)) != NULL)
      {
	WBAVLDuplicates(avltree,WB_TRUE);
	switch(listinfo->sortby)
	  {
	    case SORTDATE:
	      WBAVLCompareF(avltree,listinfo,(int (*)())SortByDate);
	    break;
	    case SORTTIME:
	      WBAVLCompareF(avltree,listinfo,(int (*)())SortByTime);
	    break;
	    case SORTSIZE:
	      WBAVLCompareF(avltree,listinfo,(int (*)())SortBySize);
	    break;
	    default:
	      WBAVLCompareF(avltree,listinfo,(int (*)())SortByName);
	  }

	WBAVLExecuteF(avltree,listinfo,(void (*)())ListDir);
	WBAVLDeleteF(avltree,NULL,(void	(*)())FreeList);

	if (_dos_findfirst("*.*",0xff,&direntry) == 0)
	  do
	    {
	      if ((direntry.attrib&FA_DIREC)
		&& strcmp(direntry.name,".")
		&& strcmp(direntry.name,"..")
		&& listinfo->recursive)
		{
		  if (chdir(direntry.name) == 0)
		    {
		      GetDir(listinfo, matchstr);
		      chdir("..");
		    }
		}
	      else if (strmatch(direntry.name,matchstr))
		{
		  if ((avlentry	= (struct dir_list *) malloc(sizeof(struct dir_list))) != NULL)
		    {
		      stat(direntry.name,&filestat);
		      avlentry->attrib = direntry.attrib;
		      avlentry->mode = filestat.st_mode;
		      avlentry->date = filestat.st_atime;
		      avlentry->size = direntry.size;

		      if ((avlentry->filename =	(char *)malloc(strlen(direntry.name)+1)) != NULL)
			{
			  strcpy(avlentry->filename,direntry.name);

			  if (listinfo->maxlen < (int) strlen(direntry.name))
			    listinfo->maxlen = strlen(direntry.name);

			  WBAVLAdd(avltree,avlentry);
			}
		    }
		}
	    }
	  while	(_dos_findnext(&direntry) == 0);

	getcwd(dirstr,MAXSTR);
	strptr = dirstr;
	while (*strptr)
	  {
	    if (*strptr	== '\\')
	      *strptr =	'/';
	    strptr++;
	  }
	printf("\n\n**** %s%s%s\n",dirstr,
				   dirstr[strlen(dirstr)-1] == '/' ? ""	: "/",
				   matchstr);

	listinfo->col =	0;
	listinfo->totfiles = 0;
	listinfo->totsize = 0;

	WBAVLInOrder(avltree);

	if (listinfo->listtype != LISTLONG)
	  printf("\n");
	printf("--------------------------------------\n");
	printf("Directory: %14.14s  %5ld files\n",
		      formatlong(tmpstr,40,listinfo->totsize),
		      listinfo->totfiles);

	WBAVLClose(avltree);
      }
  }

void driveinfo(int drive)
  {
    struct diskfree_t dtable;
    char pstr[41];

    _dos_getdiskfree((unsigned char) drive, &dtable);

    if (dtable.sectors_per_cluster != 0xFFFF)
      {
	printf("-------------------------------------------------------\n");
	formatlong(pstr,40,(long)dtable.avail_clusters*(long)dtable.sectors_per_cluster*(long)dtable.bytes_per_sector);
	printf("Available: %14s",pstr);
	formatlong(pstr,40,(long)dtable.total_clusters*(long)dtable.sectors_per_cluster*(long)dtable.bytes_per_sector);
	printf("     Disk size:	%14s\n",pstr);
      }
  }

int ProcessFiles(struct	list_info *listinfo)
  {
    char curdir[MAXSTR+1], str[MAXSTR+1], *strptr;
    int	curdrive, last = 0, done, i;
    unsigned int tmpint;

    curdrive = _getdrive();
    getcwd(curdir,MAXSTR);

    if ((strptr	= WBSListGet(listinfo->namelist,WB_FIRST)) != NULL)
      do
	{
	  last = 0;
	  if (*(strptr+1) == ':')
	    {
	      _dos_setdrive(toupper(*strptr)-64,&tmpint);
	      strptr+=2;
	    }
	  if (listinfo->nameonly == 0 && strchr(strptr,'/'))
	    {
	      done = 0;
	      i	= 0;
	      if (*strptr == '/')
		i = 1;
	      while (!done && *strptr)
		{
		  while	(strptr[i] && strptr[i]	!= '/')
		    i++;
		  strncpy(str,strptr,i);
		  *(str+i) = '\0';
		  if (chdir(str) == 0)
		      strptr +=	i;
		  else done = 1;
		  if (*strptr == '/')
		    strptr++;
		}

	      if (*strptr)
		GetDir(listinfo,strptr);
	      else
		GetDir(listinfo,"*");
	    }
	  else if (listinfo->nameonly == 0 && chdir(strptr) == 0)
	    GetDir(listinfo,"*");
	  else if (strlen(strptr))
	    GetDir(listinfo,strptr);
	  else
	    GetDir(listinfo,"*");

	  if (curdrive != _getdrive())
	    {
	      if (listinfo->recursive || listinfo->totsize != listinfo->gtotsize)
		printf("   Totals: %14.14s  %5ld files\n",
				  formatlong(str,40,listinfo->gtotsize),
				  listinfo->gtotfiles);

	      driveinfo(_getdrive());
	      _dos_setdrive(curdrive,&tmpint);
	      last = 1;
	    }
	  chdir(curdir);
	}
      while ((strptr = WBSListGet(listinfo->namelist,WB_NEXT)) != NULL);
    else GetDir(listinfo,"*");

    if (last ==	0)
      {
	if (listinfo->recursive	|| listinfo->totsize !=	listinfo->gtotsize)
	  printf("   Totals: %14.14s  %5ld files\n",
			    formatlong(str,40,listinfo->gtotsize),
			    listinfo->gtotfiles);

	driveinfo(_getdrive());
      }

    return(0);
  }

void Help(struct list_info *listinfo)
  {
    if (listinfo->namelist)
      WBSListClose(listinfo->namelist);

    printf("Usage: ls [-lh?xrRs(dts)] [file [file [...]]]\n\n");
    printf("Where: l = Long listing\n");
    printf("	   x = List accross (default)\n");
    printf("	   F = Add '/' to directorys and '*' to	executables\n");
    printf("	   r = Reverse sort\n");
    printf("	   R = Recursive listing\n");
    printf("	   s = Sort by	  (default is sort by name)\n");
    printf("	       d = Date	and Time\n");
    printf("	       t = Time\n");
    printf("	       s = Size\n");
    printf("	   d = List name only (do not recurse into directory)\n");
    printf("	   D = Disk Usage (do not print	files)\n");
    printf("	   h = This help\n");
    printf("	   ? = This help\n\n");
    printf("Unlike MS-DOS, ls uses true	wildcard characters ('*','?'), therefore,\n");
    printf("*.*	will only work if there	is a '.' in the	filename.  Futhermore,\n");
    printf("multiple '*' and '?' can be	used to	make a string match.\n\n");
    printf("	       Examples:\n");
    printf("			 ls *  - will list every file\n");
    printf("			 ls *.c	- will list every file ending with '.c'\n");
    printf("			 ls *str*c - will list every file ending with 'c'\n");
    printf("				     and containing 'str' somewhere in the\n");
    printf("				     string.\n\n");
    printf("File attributes are	defined	as:\n\n");
    printf("	     d------  File is a	directory\n");
    printf("	     -n-----  File is a	normal file\n");
    printf("	     --r----  File is read only\n");
    printf("	     ---h---  File is hidden from normal directory searches\n");
    printf("	     ----s--  File is an MS-DOS	system file\n");
    printf("	     -----l-  File is a	directory label\n");
    printf("	     ------a  File is an archive\n");
    exit(1);
  }

void ProcessArgs(int argc, char	**argv,	struct list_info *listinfo)
  {
    char *strptr;
    int	i, i2;

    listinfo->listtype = SORTACCROSS;

    if ((listinfo->namelist = WBSListOpen(NULL)) != NULL)
      {
	for (i = 1; i <	argc; i++)
	  {
	    if (argv[i][0] == '-')
	      {
		for (i2	= 1; strchr("- ",argv[i][i2]) == 0 && argv[i][i2] != 0;	i2++)
		  switch(argv[i][i2])
		    {
		      case 'h':
		      case '?':
			Help(listinfo);
		      break;
		      case 'd':	listinfo->nameonly = 1;	break;
		      case 'D':	listinfo->diskusage = 1; break;
		      case 'l':	listinfo->listtype = LISTLONG; break;
		      case 'x':	listinfo->listtype = SORTACCROSS; break;
		      case 'F':	listinfo->charname = 1;	break;
		      case 'r':	listinfo->reverse = 1; break;
		      case 'R':	listinfo->recursive = 1; break;
		      case 's':
			switch(argv[i][++i2])
			  {
			    case 'd': listinfo->sortby = SORTDATE; break;
			    case 't': listinfo->sortby = SORTTIME; break;
			    case 's': listinfo->sortby = SORTSIZE; break;
			    default:
			      printf("**** Invalid sort	by option - %s\n",argv[i]);
			      Help(listinfo);
			  }
		      break;
		      default:
			printf("**** Invalid option - %s\n",argv[i]);
			Help(listinfo);
		    }
		}
	      else
		{
		  strupr(argv[i]);
		  strptr = argv[i];
		  while	(*strptr)
		    {
		      if (*strptr == '\\')
			*strptr	= '/';
		      strptr++;
		    }
		  WBSListAdd(listinfo->namelist,argv[i],WB_LAST);
		}
	  }
      }
  }

int main(int argc, char	**argv)
  {
    struct list_info listinfo;

    memset(&listinfo,0,sizeof(struct list_info));

    ProcessArgs(argc,argv,&listinfo);
    ProcessFiles(&listinfo);

    if (listinfo.namelist)
      WBSListClose(listinfo.namelist);

    return(0);
  }
