/*****************************************************************************\
**                                                                           **
**  WW     WW IIIIIIII NNN   NN DDDDDDD  BBBBBBB     AA     SSSSSS EEEEEEEE  **
**  WW  W  WW    II    NNNN  NN DD    DD BB    BB  AA  AA  SS      EE        **
**  WW  W  WW    II    NN NN NN DD    DD BBBBBBB  AAAAAAAA  SSSSSS EEEEEE    **
**   WW W WW     II    NN  NNNN DD    DD BB    BB AA    AA      SS EE        **
**    WWWWW   IIIIIIII NN   NNN DDDDDDD  BBBBBBB  AA    AA SSSSSS  EEEEEEEE  **
**                                                                           **
**   SSSSSS  OOOOOO  FFFFFFFF TTTTTTTT WW     WW    AA    RRRRRRR  EEEEEEEE  **
**  SS      OO    OO FF          TT    WW  W  WW  AA  AA  RR    RR EE        **
**   SSSSS  OO    OO FFFFF       TT    WW  W  WW AAAAAAAA RRRRRRR  EEEEEE    **
**       SS OO    OO FF          TT     WW W WW  AA    AA RR   RR  EE        **
**  SSSSSS   OOOOOO  FF          TT      WWWWW   AA    AA RR    RR EEEEEEEE  **
**                                                                           **
*********** NOTICE ************************************************************
**        This file contains valuable trade secrets and proprietary          **
**        assets of Windbase Software Inc.  Embodying substantial            **
**        creative efforts and confidential information.  Unauthorized       **
**        use, copying, decompiling, translating, disclosure or              **
**        transfer, of any kind, is strictly prohibited.                     **
**                                                                           **
**        COPYRIGHT (C) 1992, 1993, 1994.  Windbase Software Inc.            **
**        ALL RIGHTS RESERVED.                                               **
\*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>

#include "../../memsl.h"
#include "vedt.h"

/*************************************************************************\
**  HideBlock() sets the block pointers ->p1 and ->p2 to NULL.
**  HideBlock() also sets the hilight char for all lines between ->p1
**  and ->p2 inclusive to zero.  HideBlock() redraws the entire screen
**  during the next screen update.
\*************************************************************************/
void HideBlock(EDITOR *editor)
  {
    editor->hideblock = editor->hideblock ? 0 : 1;
    UpdateCoordinates(editor,0,MAXROW-1);
  }

void CloseBlock(EDITOR *editor)
  {
    if (editor->p1 && editor->p2)
      {
        WBDListClose(editor->p1);
        WBDListClose(editor->p2);
        editor->p1 = editor->p2 = (WBDLIST *) 0;
        editor->p1pos = editor->p2pos = 0;
        editor->p1line = editor->p2line = 0;
        editor->hideblock = 0;
        UpdateCoordinates(editor,0,MAXROW-1);
      }
  }

unsigned int ChkBlkPtrs(EDITOR *editor)
  {
    unsigned int rval = 0;

    if (editor->hideblock)
      {
        editor->hideblock = 0;
        if (editor->p1 && editor->p2)
          CloseBlock(editor);
        rval = NOPTRS;
      }
    else
      {
        if (editor->p1)
          {
            if (WBDListIsPtrEqual(editor->dlist,editor->p1)
              && editor->p1pos == editor->offset+editor->pos)
              rval |= PTR1;
          }
        else rval |= NOPTR1;

        if (editor->p2)
          {
            if (WBDListIsPtrEqual(editor->dlist,editor->p2)
              && editor->p2pos == editor->offset+editor->pos)
              rval |= PTR2;
          }
        else rval |= NOPTR2;
      }

    return(rval);
  }

void SetBlkPtrs(EDITOR *editor, unsigned int ptrbits)
  {
    WBDLIST *tmpdlist;
    long tmpline;
    int tmppos;

    if (ptrbits&PTR1)
      {
        WBDListSetPtrEqual(editor->p1,editor->dlist);
        editor->p1pos = editor->offset+editor->pos;
        editor->p1line = editor->linenum;
        UpdateCoordinates(editor,0,MAXROW-1);
      }

    if (ptrbits&PTR2)
      {
        WBDListSetPtrEqual(editor->p2,editor->dlist);
        editor->p2pos = editor->offset+editor->pos;
        editor->p2line = editor->linenum;
        UpdateCoordinates(editor,0,MAXROW-1);
      }

    if (WBDListIsPtrEqual(editor->p1,editor->p2))
      {
        if (editor->p1pos == editor->p2pos)
          CloseBlock(editor);
        else if (editor->p1pos > editor->p2pos)
          {
            tmppos = editor->p2pos;
            editor->p2pos = editor->p1pos;
            editor->p1pos = tmppos;
          }
      }
    else if (editor->p2line < editor->p1line)
      {
        tmpdlist = editor->p2;
        editor->p2 = editor->p1;
        editor->p1 = tmpdlist;
        tmpline = editor->p2line;
        editor->p2line = editor->p1line;
        editor->p1line = tmpline;
        tmppos = editor->p2pos;
        editor->p2pos = editor->p1pos;
        editor->p1pos = tmppos;
      }
  }


void BeginBlock(EDITOR *editor)
  {
    editor->hideblock = 0;
    if (editor->p1)
      SetBlkPtrs(editor,PTR1);
    else
      SetBlock(editor);
    UpdateCoordinates(editor,0,MAXROW-1);
  }

void EndBlock(EDITOR *editor)
  {
    editor->hideblock = 0;
    if (editor->p2)
      SetBlkPtrs(editor,PTR2);
    else
      SetBlock(editor);
    UpdateCoordinates(editor,0,MAXROW-1);
  }

/*************************************************************************\
**  SetBlock() is used by several of the blocking routines and simply
**  initailizes the blocking pointers ->p1 and ->p2.
\*************************************************************************/
void SetBlock(EDITOR *editor)
  {
    if ((editor->p1 = WBDListPointer(editor->dlist)) != NULL)
      if ((editor->p2 = WBDListPointer(editor->dlist)) != NULL)
        {
          editor->p1pos = editor->p2pos = editor->offset + editor->pos;
          editor->p1line = editor->p2line = editor->linenum;

        }
      else
        {
          WBDListClose(editor->p1);
          editor->p1 = (WBDLIST *) 0;
        }
  }

int BeginningOfBlk(EDITOR *editor)
  {
    if (editor->hideblock == 0 && editor->p1)
      {
        if (editor->changed)
          UpdateCurrent(editor);

        MoveToLine(editor,editor->p1);
        editor->pos = editor->p1pos;
        UpdateCoordinates(editor,editor->row,MAXROW-1);
        LineMove(editor);
        return(1);
      }
    return(0);
  }

int EndOfBlk(EDITOR *editor)
  {
    if (editor->hideblock == 0 && editor->p2)
      {
        if (editor->changed)
          UpdateCurrent(editor);

        MoveToLine(editor,editor->p2);
        editor->pos = editor->p2pos;
        UpdateCoordinates(editor,editor->row,MAXROW-1);
        LineMove(editor);
        return(1);
      }
    return(0);
  }

/*************************************************************************\
**  If ->p1 and ->p2 are not equal then BlockCopy() will copy all the
**  lines between ->p1, ->p1pos and ->p2, p2pos and will free and set
**  charbuffer to NULL.
**  if ->p1 and ->p2 are equal, but ->p1pos and ->p2pos are not then
**  BlockCopy() will copy the chars between ->p1pos and ->p2pos to the
**  character buffer (charbuffer) and close and set dlistbuffer to NULL.
\*************************************************************************/
void BlockCopy(EDITOR *editor)
  {
    char *strptr, *tmpstr;

    if (editor->hideblock == 0 && editor->p1 && editor->p2)
      {
        if (editor->dlistbuffer)
          WBDListDelete(editor->dlistbuffer,WB_CLEAR);

        if (editor->changed)
          UpdateCurrent(editor);

        if (editor->charbuffer)
          *editor->charbuffer = '\0';

        if (WBDListIsPtrEqual(editor->p1,editor->p2))
          {
            if (editor->charbuffer)
              if ((strptr = WBDListGet(editor->p1,WB_CURRENT)) != NULL)
                {
                  strncpy(editor->charbuffer,strptr+1+editor->p1pos,editor->p2pos-editor->p1pos);
                  *(editor->charbuffer+editor->p2pos-editor->p1pos) = '\0';
                }
          }
        else if (editor->dlistbuffer)
          {
            if (WBDListCopy(editor->dlistbuffer,editor->p1,editor->p2) != 0)
              {
                if (editor->p1pos)
                  if ((strptr = WBDListGet(editor->dlistbuffer,WB_FIRST)) != NULL)
                    if (editor->p1pos < (int) strlen(strptr+1))
                      {
                        strcpy(strptr+1,strptr+1+editor->p1pos);
                        UpdateLine(editor->dlistbuffer,strptr,1);
                      }

                if (editor->p2pos)
                  {
                    if ((strptr = WBDListGet(editor->dlistbuffer,WB_LAST)) != NULL)
                      if (editor->p2pos < (int) strlen(strptr+1))
                        {
                          *(strptr+1+editor->p2pos) = '\0';
                          UpdateLine(editor->dlistbuffer,strptr,1);
                        }
                  }
                else
                  {
                    WBDListDelete(editor->dlistbuffer,WB_LAST);
                    strptr = WBDListGet(editor->dlistbuffer,WB_LAST);
                    if ((tmpstr = (char *)malloc(strlen(strptr+1)+3)) != NULL)
                      {
                        strcpy(tmpstr+1,strptr+1);
                        strcat(tmpstr+1,"\n");
                        WBDListAdd(editor->dlistbuffer,tmpstr,WB_CURRENT);
                      }
                  }
              }
            else WBDListDelete(editor->dlistbuffer,WB_CLEAR);
          }
      }
  }

/*************************************************************************\
**  BlockCut() is similar to BlockCopy except that it actually moves the
**  block to the buffer.
\*************************************************************************/
void BlockCut(EDITOR *editor)
  {
    WBDLIST *cutbuffer;
    char *strptr, *strptr2, *tmpptr, *tmpstr = (char *) 0;
    int settop = 0;

    if (editor->hideblock == 0 && editor->p1 && editor->p2)
      {
        if (editor->dlistbuffer)
          WBDListDelete(editor->dlistbuffer,WB_CLEAR);

        if (editor->charbuffer)
          *editor->charbuffer = '\0';

        if (editor->changed)
          UpdateCurrent(editor);

        if (WBDListIsPtrEqual(editor->p1,editor->dlist) == 0)
          MoveToLine(editor,editor->p1);

        editor->pos = editor->p1pos;

        if (WBDListIsPtrEqual(editor->p1,editor->p2))
          {
            if (editor->charbuffer)
              {
                strncpy(editor->charbuffer,editor->str+1+editor->p1pos,editor->p2pos-editor->p1pos);
                *(editor->charbuffer+editor->p2pos-editor->p1pos) = '\0';
                strcpy(editor->str+1+editor->p1pos,editor->str+1+editor->p2pos);
                editor->changed = editor->realloc = 1;
                UpdateCoordinates(editor,editor->row,editor->row);
              }
          }
        else if (editor->dlistbuffer)
          {
            if (editor->p1pos)
              if ((strptr = WBDListGet(editor->p1,WB_CURRENT)) != NULL)
                if (editor->p1pos < (int) strlen(strptr+1))
                  if ((tmpstr = malloc(strlen(strptr+1+editor->p1pos)+2)) != NULL)
                    {
                      strcpy(tmpstr+1,strptr+1+editor->p1pos);
                      *(strptr+1+editor->p1pos) = '\0';
                      WBDListAdd(editor->dlistbuffer,tmpstr,WB_LAST);
                      tmpstr = (char *) 0;
                    }

            strptr2 = WBDListGet(editor->p2,WB_CURRENT);

            if (editor->p2pos)
              if ((tmpstr = malloc(editor->p2pos+2)) != NULL)
                {
                  strncpy(tmpstr+1,strptr2+1,editor->p2pos);
                  *(tmpstr+1+editor->p2pos) = '\0';
		  if (editor->p2pos < strlen(strptr2+1))
		    strcpy(strptr2+1,strptr2+1+editor->p2pos);
		  else
		    *(strptr2+1) = '\0';
                }

            if (editor->p1pos)
              {
                if (strlen(strptr+1)+strlen(strptr2+1) < MAXSTR)
                  {
                    if ((tmpptr = malloc(strlen(strptr+1)+strlen(strptr2+1)+2)) != NULL)
                      {
                        strcpy(tmpptr+1,strptr+1);
                        strcat(tmpptr+1,strptr2+1);
                        UpdateLine(editor->p1,tmpptr,1);
                        WBDListGet(editor->p1,WB_NEXT);
                        WBDListDelete(editor->p2,WB_CURRENT);
                      }
                  }
              }

            if (WBDListIsPtrEqual(editor->p1,editor->p2) == 0)
              {
                WBDListGet(editor->p2,WB_PREV);

                if (WBDListIsPtrEqual(editor->p1,editor->dltopscr))
                  settop = 1;

                if ((cutbuffer = WBDListOpen(NULL)) != NULL)
                  {
                    if (WBDListCut(cutbuffer,editor->p1,editor->p2))
                      WBDListAddList(editor->dlistbuffer,cutbuffer,WB_LAST,0);
                    WBDListClose(cutbuffer);
                  }

                if (editor->p1pos)
                  WBDListGet(editor->dlist,WB_PREV);

                if (settop)
                  WBDListSetPtrEqual(editor->dltopscr,editor->dlist);
              }

            if (tmpstr)
              WBDListAdd(editor->dlistbuffer,tmpstr,WB_LAST);
            else
              {
                strptr = WBDListGet(editor->dlistbuffer,WB_LAST);
                if ((tmpstr = (char *)malloc(strlen(strptr+1)+3)) != NULL)
                  {
                    strcpy(tmpstr+1,strptr+1);
                    strcat(tmpstr+1,"\n");
                    WBDListAdd(editor->dlistbuffer,tmpstr,WB_CURRENT);
                  }
              }

            UpdateCoordinates(editor,editor->row,MAXROW-1);
            editor->chgmade = 1;
            LineMove(editor);
          }
        CloseBlock(editor);
      }
  }

/*************************************************************************\
**
\*************************************************************************/
void BlockInsert(EDITOR *editor)
  {
    WBDLIST *pointer;
    char *strptr, *strptr2, *tmpstr;
    long numlines;
    int settop = 0;

    if (editor->changed)
      UpdateCurrent(editor);

    CloseBlock(editor);

    if (editor->charbuffer && *editor->charbuffer)
      {
	if (strlen(editor->str+1) + strlen(editor->charbuffer) < MAXSTR-1)
	  {
	    SetBlock(editor);
	    if (editor->pos + editor->offset > strlen(editor->str+1))
	      memset(editor->str+1+strlen(editor->str+1),' ',(editor->pos+editor->offset) - strlen(editor->str+1));
            insstr(editor->str+1,editor->charbuffer,editor->offset+editor->pos);
            editor->p2pos+=strlen(editor->charbuffer);
            editor->changed = editor->realloc = 1;
            UpdateCoordinates(editor,editor->row,editor->row);
          }
      }
    else if (editor->dlistbuffer
      && (WBDListIs(editor->dlistbuffer)&WB_EMPTY) != WB_EMPTY)
      {
        if (WBDListIsPtrEqual(editor->dlist,editor->dltopscr))
          settop = 1;

        numlines = WBDListNumItems(editor->dlist);

        pointer = WBDListPointer(editor->dlist);
        WBDListDeleteF(pointer,NULL,(void (*)())DelFunct);
        if (WBDListAddList(editor->dlist,editor->dlistbuffer,WB_PREV,1) != 0)
          {
            if (settop)
              WBDListSetPtrEqual(editor->dltopscr,editor->dlist);

            SetBlock(editor);
            WBDListSetPtrEqual(editor->p1,editor->dlist);
            editor->p1line = editor->linenum;
            editor->p1pos = 0;

            if (editor->pos+editor->offset > 0)
              {
                if (pointer && (strptr = WBDListGet(pointer,WB_CURRENT)) != NULL)
                  if ((strptr2 = WBDListGet(editor->dlist,WB_CURRENT)) != NULL)
                  {
                    if ((tmpstr = malloc(strlen(strptr2+1)+editor->pos+editor->offset+2)) != NULL)
                      {
                        strncpy(tmpstr+1,strptr+1,editor->pos+editor->offset);
                        *(tmpstr+1+editor->pos+editor->offset) = '\0';
                        if ((int) strlen(tmpstr+1) < editor->pos+editor->offset)
                          memset(tmpstr+1+strlen(tmpstr+1),' ',editor->pos+editor->offset-strlen(tmpstr+1));
                        strcat(tmpstr+1,strptr2+1);
                        WBDListAdd(editor->dlist,tmpstr,WB_CURRENT);
                        if ((int) strlen(strptr+1) < editor->pos+editor->offset)
                          *(strptr+1) = '\0';
                        else
                          strcpy(strptr+1,strptr+1+editor->pos+editor->offset);
                        UpdateLine(pointer,strptr,1);
                        editor->p1pos = editor->pos+editor->offset;
                      }
                  }
              }

            strptr2 = WBDListGet(pointer,WB_CURRENT);
            strptr = WBDListGet(pointer,WB_PREV);
            if (*strptr+1 && *(strptr+strlen(strptr+1)) == '\n')
              {
                *(strptr+strlen(strptr+1)) = '\0';
                UpdateLine(pointer,strptr,1);
                WBDListGet(pointer,WB_NEXT);
                editor->p2pos = 0;
              }
            else
              {
                if (strlen(strptr+1) + strlen(strptr2+1) < MAXSTR-1)
                  {
                    if ((tmpstr = malloc(strlen(strptr+1)+strlen(strptr2+1)+2)) != NULL)
                      {
			editor->p2pos = strlen(strptr+1);
			strcpy(tmpstr+1,strptr+1);
			strcat(tmpstr+1,strptr2+1);
			UpdateLine(pointer,tmpstr,1);
			WBDListDelete(pointer,WB_NEXT);
			WBDListGet(pointer,WB_PREV);
		      }
                  }
              }

            editor->chgmade = 1;
            LineMove(editor);
            UpdateCoordinates(editor,editor->row,MAXROW-1);
            WBDListSetPtrEqual(editor->p2,pointer);
            editor->p2line = editor->p1line
                           + (WBDListNumItems(editor->dlist) - numlines);
          }
      }
  }

void BlockUprLwr(EDITOR *editor, int upper)
  {
    WBDLIST *dlistptr;
    char *strptr;
    int i, done = 0;

    if (editor->p1 && editor->p2)
      {
        if (editor->changed)
          UpdateCurrent(editor);
        if (WBDListIsPtrEqual(editor->p1,editor->p2))
          {
            for (i = editor->p1pos; i < editor->p2pos; i++)
              if (upper)
                *(editor->str+i+1) = *(editor->strptr+i+1) = (char) toupper(*(editor->str+i+1));
              else
                *(editor->str+i+1) = *(editor->strptr+i+1) = (char) tolower(*(editor->str+i+1));
          }
        else
          {
            if ((dlistptr = WBDListPointer(editor->p1)) != NULL)
              {
                strptr = WBDListGet(dlistptr,WB_CURRENT);
                while (! done && strptr)
                  {
                    if (WBDListIsPtrEqual(dlistptr,editor->p1))
                      if (upper)
                        strupr(strptr+editor->p1pos+1);
                      else
                        strlwr(strptr+editor->p1pos+1);
                    else if (WBDListIsPtrEqual(dlistptr,editor->p2))
                      {
                        for (i = 1; i < editor->p2pos+1; i++)
                          if (upper)
                            *(strptr+i) = (char) toupper(*(strptr+i));
                          else
                            *(strptr+i) = (char) tolower(*(strptr+i));
                      }
                    else
                      {
                        if (upper)
                          strupr(strptr+1);
                        else
                          strlwr(strptr+1);
                      }
                    if (WBDListIsPtrEqual(dlistptr,editor->p2))
                      done = 1;
                    else
                      strptr = WBDListGet(dlistptr,WB_NEXT);
                  }
                WBDListClose(dlistptr);
              }
          }
        editor->chgmade = 1;
        LineMove(editor);
        UpdateCoordinates(editor,0,MAXROW-1);
      }
  }
