/*****************************************************************************\
**                                                                           **
**  WW     WW IIIIIIII NNN   NN DDDDDDD  BBBBBBB     AA     SSSSSS EEEEEEEE  **
**  WW  W  WW    II    NNNN  NN DD    DD BB    BB  AA  AA  SS      EE        **
**  WW  W  WW    II    NN NN NN DD    DD BBBBBBB  AAAAAAAA  SSSSSS EEEEEE    **
**   WW W WW     II    NN  NNNN DD    DD BB    BB AA    AA      SS EE        **
**    WWWWW   IIIIIIII NN   NNN DDDDDDD  BBBBBBB  AA    AA SSSSSS  EEEEEEEE  **
**                                                                           **
**   SSSSSS  OOOOOO  FFFFFFFF TTTTTTTT WW     WW    AA    RRRRRRR  EEEEEEEE  **
**  SS      OO    OO FF          TT    WW  W  WW  AA  AA  RR    RR EE        **
**   SSSSS  OO    OO FFFFF       TT    WW  W  WW AAAAAAAA RRRRRRR  EEEEEE    **
**       SS OO    OO FF          TT     WW W WW  AA    AA RR   RR  EE        **
**  SSSSSS   OOOOOO  FF          TT      WWWWW   AA    AA RR    RR EEEEEEEE  **
**                                                                           **
*********** NOTICE ************************************************************
**        This file contains valuable trade secrets and proprietary          **
**        assets of Windbase Software Inc.  Embodying substantial            **
**        creative efforts and confidential information.  Unauthorized       **
**        use, copying, decompiling, translating, disclosure or              **
**        transfer, of any kind, is strictly prohibited.                     **
**                                                                           **
**        COPYRIGHT (C) 1992, 1993, 1994.  Windbase Software Inc.            **
**        ALL RIGHTS RESERVED.                                               **
\*****************************************************************************/

#include <conio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>

#include "../../memsl.h"
#include "vedt.h"
#include "getkey.h"

/*************************************************************************\
**  KeyStat() is used to retrieve the current keyboard control key status.
**  The check is made with the low level DOS call 0x16 (Keyboard services)
**  with 'ah' set to 2.  The DOS call can also be used to check the status
**  of other control keys.  Since all <CTRL> and <ALT> key combinations
**  generate a key sequence under MS-DOS, we are only interested in the
**  <SHIFT> key.
\*************************************************************************/

void KeyStat(struct key_stat *keystat)
  {
    union REGS regs;

    memset(keystat,0,sizeof(struct key_stat));
    memset(&regs,0,sizeof(union REGS));

    regs.h.ah = 02;

    int86(0x16,&regs,&regs);

    if ((regs.h.al&0x80) == 0x80)
      keystat->insert = 1;
    if ((regs.h.al&0x01) == 0x01)
      keystat->shiftkey = 1;
    if ((regs.h.al&0x02) == 0x02)
      keystat->shiftkey = 1;
  }

/*************************************************************************\
**  GetKey() is the main keyboard processor.  GetKey() loops continuously
**  until all keys are gathered from the keyboard and the keyboard is
**  idle.  When the keyboard is idle GetKey() will return a character with
**  no delay.  GetKey() translates the keypresses into the key codes as
**  as defined in getkey.h.  If the queue is full additional key presses
**  are ignored.
\*************************************************************************/

int GetKey(EDITOR *editor)
  {
    int ch, skip;
    struct key_stat keystat;

      /*
      ** If the queue is full the key press is ignored for the moment.
      ** If a key has been pressed and the queue is not full, process
      ** the key.  If no keys were pressed and the queue is empty,
      ** then there are no keys to return and GetKey() simply waits
      ** for another key to be pressed.
      */
    while (WBQueueNumItems(editor->keybuff) < WBQueueMaxItems(editor->keybuff)
      && (kbhit() || WBQueueNumItems(editor->keybuff) == 0))
      {
	skip = 0;
	switch(ch = getch())
	  {
	    case 0:
		/*
		** DOS returns zero along with a key code for special keys.
		** For example, if the <F1> key is pressed, zero and ';'
		** will be sitting in the key pool.
		*/
	      switch(ch = getch())
		{
		  case 59: ch = KEY_HELP; break;      /* F1 */
		  case 60: ch = KEY_LOAD; break;      /* F2 */
		  case 61: ch = KEY_SEARCH; break;    /* F3 */
		  case 62: ch = KEY_SRCHNEXT; break;  /* F4 */
		  case 63: ch = KEY_MACRO; break;     /* F5 */
		  case 64: ch = KEY_MACRORD; break;   /* F6 */
		  case 67: ch = KEY_READBLK; break;   /* F9 */
		  case 85: ch = KEY_SAVE; break;      /* SF2 */
		  case 86: ch = KEY_SRCHRPLC; break;  /* SF3 */
		  case 87: ch = KEY_SRCHNEXTBK; break;/* SF4 */
		  case 88: ch = KEY_MACROEXE; break;  /* SF5 */
		  case 89: ch = KEY_MACROWR; break;   /* SF6 */
		  case 92: ch = KEY_WRITEBLK; break;  /* SF9 */
		  case 95: ch = KEY_SAVEAS; break;    /* CF2 */
		  case 98: ch = KEY_MACROBLK; break;  /* CF5 */
		  case 102: ch = KEY_HIDEBLOCK; break;/* CF9 */
		  case 71:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SHOME;
		    else
		      ch = KEY_HOME;
		  break;
		  case 72:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SUP;
		    else
		      ch = KEY_UP;
		  break;
		  case 73:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SPGUP;
		    else
		      ch = KEY_PGUP;
		  break;
		  case 75:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SLEFT;
		    else
		      ch = KEY_LEFT;
		  break;
		  case 77:
		    KeyStat(&keystat);
                    if (keystat.shiftkey)
                      ch = KEY_SRIGHT;
                    else
		      ch = KEY_RIGHT;
		  break;
                  case 79:
		    KeyStat(&keystat);
                    if (keystat.shiftkey)
		      ch = KEY_SEND;
                    else
                      ch = KEY_END;
		  break;
		  case 80:
                    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SDN;
                    else
                      ch = KEY_DN;
		  break;
                  case 81:
                    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SPGDN;
		    else
                      ch = KEY_PGDN;
                  break;
		  case 82:
                    KeyStat(&keystat);
                    if (keystat.shiftkey)
                      ch = KEY_SINSERT;
		    else
		      ch = KEY_INSERT;
                  break;
		  case 83:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SDELETE;
		    else
		      ch = KEY_DELETE;
		  break;
		  case 115:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SCLEFT;
		    else
		      ch = KEY_CLEFT;
		  break;
		  case 116:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SCRIGHT;
		    else
		      ch = KEY_CRIGHT;
		  break;
		  case 117: ch = KEY_CEND; break;
		  case 118:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SCPGDN;
		    else
		      ch = KEY_CPGDN;
		  break;
		  case 119: ch = KEY_CHOME; break;
		  case 132:
		    KeyStat(&keystat);
		    if (keystat.shiftkey)
		      ch = KEY_SCPGUP;
		    else
		      ch = KEY_CPGUP;
		  break;
		  case 141: ch = KEY_CUP; break;
		  case 145: ch = KEY_CDN; break;
		  case 146: ch = KEY_CINSERT; break;
		  case 147: ch = KEY_CDELETE; break;
		  case 162: ch = KEY_AINSERT; break;
		  case 163: ch = KEY_ADELETE; break;
		  default:
		    /* Key unknown - skip */
		    skip = 1;
		}
	    break;
	    case 8: ch = KEY_BACKSPACE; break;
	    case 9: ch = KEY_TAB; break;
	    case KEY_CTRLB:
	      if ((ch = getch()) < 27)
		if (ch != 0)
		  ch += 96;
	      ch = toupper(ch);
	      switch(ch)
		{
		  case 0: skip = 1; ch = getch(); break;
		  case 'E': ch = KEY_EXIT; break;
		  case 'H': ch = KEY_HELP; break;
		  case 'L': ch = KEY_LOAD; break;
		  case 'S': ch = KEY_SAVE; break;
		  case 'F': ch = KEY_SAVEAS; break;
		  case 'N': ch = KEY_NEXTFILE; break;
		  case 'P': ch = KEY_PREVFILE; break;
		  default: skip = 1;
		}
	    break;
	    case KEY_CTRLK:
	      if ((ch = getch()) < 27)
		if (ch != 0)
		  ch += 96;
	      ch = toupper(ch);
	      switch(ch)
		{
		  case 0: skip = 1; ch = getch(); break;
		  case 'B': ch = KEY_BEGINBLOCK; break;
		  case 'C': ch = KEY_BLOCKCPY; break;
		  case 'K': ch = KEY_ENDBLOCK; break;
		  case 'H': ch = KEY_HIDEBLOCK; break;
		  case 'L': ch = KEY_BLOCKLWR; break;
		  case 'P': ch = KEY_PASTEBLK; break;
		  case 'R': ch = KEY_READBLK; break;
		  case 'U': ch = KEY_BLOCKUPR; break;
		  case 'Y': ch = KEY_BLOCKDEL; break;
		  case 'W': ch = KEY_WRITEBLK; break;
		  default: skip = 1;
		}
	    break;
	    case KEY_CTRLO:
	      if ((ch = getch()) < 27)
		if (ch != 0)
		  ch += 96;
	      ch = toupper(ch);
	      switch(ch)
		{
		  case 0: skip = 1; ch = getch(); break;
		  case 'I': ch = KEY_AINDENT; break;
		  case 'S': ch = KEY_SRCHBACK; break;
		  default: skip = 1;
		}
	    break;
	    case KEY_CTRLP:
	      if ((ch = getch()) < 27)
		if (ch != 0)
		  ch += 96;
	      ch = toupper(ch);
	      switch(ch)
		{
		  case 0: skip = 1; ch = getch(); break;
		  case 'S': ch = KEY_MACRO; break;
		  case 'M': ch = KEY_MACROEXE; break;
		  case 'B': ch = KEY_MACROBLK; break;
		  case 'R': ch = KEY_MACRORD; break;
		  case 'W': ch = KEY_MACROWR; break;
		  default: skip = 1;
		}
	    break;
	    case KEY_CTRLQ:
	      if ((ch = getch()) < 27)
		if (ch != 0)
		  ch += 96;
	      ch = toupper(ch);
	      switch(ch)
		{
		  case 0: skip = 1; ch = getch(); break;
		  case 'A': ch = KEY_SRCHRPLC; break;
		  case 'B': ch = KEY_BGNOFBLK; break;
		  case 'D': ch = KEY_END; break;
		  case 'F': ch = KEY_SEARCH; break;
		  case 'K': ch = KEY_ENDOFBLK; break;
		  case 'L': ch = KEY_GOTOLINE; break;
		  case 'S': ch = KEY_HOME; break;
		  case 'R': ch = KEY_CPGUP; break;
		  case 'C': ch = KEY_CPGDN; break;
		  case 'E': ch = KEY_TOW; break;
		  case 'X': ch = KEY_BOW; break;
		  case 'Y': ch = KEY_DELEOL; break;
		  default: skip = 1;
		}
	    break;
	    case 13: ch = KEY_ENTER; break;
	    case 27: ch = KEY_EXIT; break;
	  }
	if (skip == 0)
	  WBQueuePush(editor->keybuff,&ch);  /* Push the key on to the queue. */
      }
    WBQueuePop(editor->keybuff,&ch);  /* Once GetKey() is out of the key */
				      /* gathering loop it will return   */
				      /* the first key in the queue.     */
    return(ch);
  }
