/*****************************************************************************\
**                                                                           **
**  WW     WW IIIIIIII NNN   NN DDDDDDD  BBBBBBB     AA     SSSSSS EEEEEEEE  **
**  WW  W  WW    II    NNNN  NN DD    DD BB    BB  AA  AA  SS      EE        **
**  WW  W  WW    II    NN NN NN DD    DD BBBBBBB  AAAAAAAA  SSSSSS EEEEEE    **
**   WW W WW     II    NN  NNNN DD    DD BB    BB AA    AA      SS EE        **
**    WWWWW   IIIIIIII NN   NNN DDDDDDD  BBBBBBB  AA    AA SSSSSS  EEEEEEEE  **
**                                                                           **
**   SSSSSS  OOOOOO  FFFFFFFF TTTTTTTT WW     WW    AA    RRRRRRR  EEEEEEEE  **
**  SS      OO    OO FF          TT    WW  W  WW  AA  AA  RR    RR EE        **
**   SSSSS  OO    OO FFFFF       TT    WW  W  WW AAAAAAAA RRRRRRR  EEEEEE    **
**       SS OO    OO FF          TT     WW W WW  AA    AA RR   RR  EE        **
**  SSSSSS   OOOOOO  FF          TT      WWWWW   AA    AA RR    RR EEEEEEEE  **
**                                                                           **
*********** NOTICE ************************************************************
**        This file contains valuable trade secrets and proprietary          **
**        assets of Windbase Software Inc.  Embodying substantial            **
**        creative efforts and confidential information.  Unauthorized       **
**        use, copying, decompiling, translating, disclosure or              **
**        transfer, of any kind, is strictly prohibited.                     **
**                                                                           **
**        COPYRIGHT (C) 1992, 1993, 1994.  Windbase Software Inc.            **
**        ALL RIGHTS RESERVED.                                               **
\*****************************************************************************/

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>

#include "../../memsl.h"

#include "port.h"
#include "vedt.h"
#include "getkey.h"

EDITOR *EditorOpen(char *filename, int viewonly, int tabspec, int entab)
  {
    EDITOR *editor;
    char *strptr;

    if ((editor = (EDITOR *) malloc(sizeof(EDITOR))) != (EDITOR *) 0)
      {
        memset(editor,0,sizeof(EDITOR));

        strptr = filename;
        while (*strptr)
          {
            if (*strptr == '\\')
              *strptr = '/';
            strptr++;
          }
        strcpy(editor->filename,filename);

        editor->viewonly = viewonly;
        editor->entab = entab;

        editor->tabspec = tabspec == 0 ? 8 : tabspec;
        InitTabs(editor->tabstr,MAXSTR,editor->tabspec);

        if ((editor->dlist = GetFile(editor,filename)) != NULL)
          {
            WBDListGet(editor->dlist,WB_FIRST);
            strcpy(editor->delims, " ~`!@#$%^&*()-_=+\\|[{]}'\";:/?.>,<");

            if ((editor->dltopscr = WBDListPointer(editor->dlist)) != NULL)
              return(editor);

            WBDListClose(editor->dlist);
          }

        printf("\nSorry.  There was not enough memory to load %s.\n",filename);
        free(editor);
      }
    return((EDITOR *) 0);
  }

void InitEditor(EDITOR *editor)
  {
    struct key_stat keystat;

    clrscr();
    reversevideo();
    gotoxy(1,24); clreol();
    gotoxy(1,25); clreol();
    normvideo();

    if (editor->viewonly == 0)
      {
        KeyStat(&keystat);
        if (keystat.insert)
          editor->insert = 1;
      }
    else
      editor->insert = 1;

    if (*editor->srchoptions == 0)
      *editor->srchoptions |= SRCH_FORWARD;

    window(1,1,MAXSCRCOL,MAXROW);
    LineMove(editor);
    UpdateCoordinates(editor,0,MAXROW-1);
    Draw(editor);
    UpdateStats(editor,1);
  }

void EditCleanUp(EDITOR *editor)
  {
    int ch;

    if (editor->changed)
      UpdateCurrent(editor);

    if (editor->p1 && editor->p2)
      {
        CloseBlock(editor);
        Draw(editor);
      }

    Cursor(2);
    window(1,1,MAXSCRCOL,MAXSCRROW);
    normvideo();
    gotoxy(1,24); clreol();
    gotoxy(1,25); clreol();

    if (editor->chgmade)
      {
        gotoxy(1,24); printf("Changes were made. Would you like to save them (y/n)? ");
        do { ch = getch(); } while (strchr("yYnN",ch) == 0);
        printf("%c\n",(char)ch);
        if (strchr("yY",ch))
          PutFile(editor,editor->dlist,editor->filename,1);
        editor->chgmade = editor->changed = editor->realloc = 0;
      }
    gotoxy(1,MAXSCRROW);
  }

void EditorClose(EDITOR *editor)
  {
    WBDListClose(editor->dltopscr);
    WBDListClose(editor->dlist);
    free(editor);
  }

void VEditHelp(EDITOR *editor, char *helpfile)
  {
    EDITOR *viewhelp;
    WBQUEUE *keybuff;
    char searchstr[MAXSRCHSTR+3];

    if (getenv("VEDTHELP"))
      helpfile = getenv("VEDTHELP");

    if ((keybuff = WBQueueOpen(NULL,sizeof(int),MAXKEYS,1)) != NULL)
      {
        if ((viewhelp = EditorOpen(helpfile,1,8,0)) != NULL)
          {
            memset(searchstr,0,MAXSRCHSTR+3);
            viewhelp->keybuff = keybuff;
            viewhelp->searchstr = searchstr;
            viewhelp->help = 1;
            VEditManager(viewhelp);
            EditorClose(viewhelp);
            InitEditor(editor);
          }
        WBQueueClose(keybuff);
      }
  }

void VEditManager(EDITOR *editor)
  {
    int rval = 0;

    Cursor(0);
    InitEditor(editor);
    do
      {
        UpdateStats(editor,0);
        gotoxy(1+editor->pos,1+editor->row);
        Cursor(editor->insert+1);
        if (editor->viewonly)
          editor->key = GetKey(editor);
        else
          ProcessKey(editor);
        Cursor(0);
        switch(editor->key)
          {
            case KEY_HELP:
              Cursor(2);
              if (editor->help == 0)
                VEditHelp(editor,"help.txt");
              Cursor(0);
            break;
            case KEY_LOAD:
              if (editor->viewonly == 0)
                {
                  Cursor(2);
                  LoadFile(editor);
                  Cursor(0);
                }
            break;
            case KEY_SAVE:
              if (editor->viewonly == 0)
                SaveFile(editor);
            break;
            case KEY_SAVEAS:
              if (editor->viewonly == 0)
                {
                  Cursor(2);
                  SaveFileAs(editor);
                  Cursor(0);
                }
            break;
            case KEY_SEARCH:
              Cursor(2);
              rval = Search(editor,1);
              Cursor(0);
            break;
            case KEY_SRCHRPLC:
              Cursor(2);
              rval = SearchReplace(editor);
              Cursor(0);
            break;
            case KEY_CTRLL:
            case KEY_SRCHNEXT:
              rval = Search(editor,0);
            break;
            case KEY_SRCHBACK:
              SearchBack(editor);
            break;
            case KEY_MACRO:
              if (editor->viewonly == 0)
                Macro(editor);
            break;
            case KEY_MACROEXE:
              if (editor->viewonly == 0)
                {
                  ExecuteMacro(editor);
                  rval = 1;
                }
            break;
            case KEY_MACROBLK:
              if (editor->viewonly == 0)
                {
                  ExecuteMacroBlock(editor);
                  rval = 1;
                }
            break;
            case KEY_MACRORD:
              if (editor->viewonly == 0)
                {
                  Cursor(2);
                  ReadMacro(editor);
                  Cursor(0);
                }
            break;
            case KEY_MACROWR:
              if (editor->viewonly == 0)
                {
                  Cursor(2);
                  WriteMacro(editor);
                  Cursor(0);
                }
            break;
            case KEY_READBLK:
              if (editor->viewonly == 0)
                {
                  Cursor(2);
                  ReadBlock(editor,NULL);
                  Cursor(0);
                }
            break;
            case KEY_WRITEBLK:
              if (editor->viewonly == 0)
                {
                  Cursor(2);
                  WriteBlock(editor);
                  Cursor(0);
                }
            break;
            case KEY_HIDEBLOCK:
              if (editor->viewonly == 0)
                HideBlock(editor);
            break;
            case KEY_HOME:
              KeyHome(editor,0);
            break;
            case KEY_SHOME:
              if (editor->viewonly)
                KeyHome(editor,0);
              else
                KeyHome(editor,1);
            break;
            case KEY_END:
              KeyEnd(editor,0);
            break;
            case KEY_SEND:
              if (editor->viewonly)
                KeyEnd(editor,0);
              else
                KeyEnd(editor,1);
            break;
            case KEY_CTRLV:
            case KEY_INSERT:
              if (editor->viewonly == 0)
                KeyInsert(editor);
            break;
            case KEY_PASTEBLK:
            case KEY_SINSERT:
              if (editor->viewonly == 0)
                BlockInsert(editor);
            break;
            case KEY_BLOCKCPY:
            case KEY_CINSERT:
              if (editor->viewonly == 0)
                BlockCopy(editor);
            break;
            case KEY_CTRLN:
            case KEY_AINSERT:
              if (editor->viewonly == 0)
                KeyInsertLine(editor);
            break;
            case KEY_ENTER:
              if (editor->viewonly == 0)
                KeyEnter(editor);
            break;
            case KEY_CTRLE:
            case KEY_UP:
              rval = KeyUp(editor,0);
            break;
            case KEY_SUP:
              if (editor->viewonly)
                rval = KeyUp(editor,0);
              else
                rval = KeyUp(editor,1);
            break;
            case KEY_CTRLR:
            case KEY_PGUP:
              rval = KeyPageUp(editor,0);
            break;
            case KEY_SPGUP:
              if (editor->viewonly)
                rval = KeyPageUp(editor,0);
              else
                rval = KeyPageUp(editor,1);
            break;
            case KEY_CTRLS:
            case KEY_LEFT:
              rval = KeyLeft(editor,0);
            break;
            case KEY_SLEFT:
              if (editor->viewonly)
                rval = KeyLeft(editor,0);
              else
                rval = KeyLeft(editor,1);
            break;
            case KEY_CTRLA:
            case KEY_CLEFT:
              rval = KeyControlLeft(editor,0);
            break;
            case KEY_SCLEFT:
              if (editor->viewonly)
                rval = KeyControlLeft(editor,0);
              else
                rval = KeyControlLeft(editor,1);
            break;
            case KEY_CTRLD:
            case KEY_RIGHT:
              rval = KeyRight(editor,0);
            break;
            case KEY_SRIGHT:
              if (editor->viewonly)
                rval = KeyRight(editor,0);
              else
                rval = KeyRight(editor,1);
            break;
            case KEY_CTRLF:
            case KEY_CRIGHT:
              rval = KeyControlRight(editor,0);
            break;
            case KEY_SCRIGHT:
              if (editor->viewonly)
                rval = KeyControlRight(editor,0);
              else
                rval = KeyControlRight(editor,1);
            break;
            case KEY_CTRLX:
            case KEY_DN:
              rval = KeyDown(editor,0);
            break;
            case KEY_SDN:
              if (editor->viewonly)
                rval = KeyDown(editor,0);
              else
                rval = KeyDown(editor,1);
            break;
            case KEY_CTRLC:
            case KEY_PGDN:
              rval = KeyPageDown(editor,0);
            break;
            case KEY_SPGDN:
              if (editor->viewonly)
                rval = KeyPageDown(editor,0);
              else
                rval = KeyPageDown(editor,1);
            break;
            case KEY_CPGDN:
              rval = KeyControlPageDown(editor,0);
            break;
            case KEY_SCPGDN:
              if (editor->viewonly)
               rval = KeyControlPageDown(editor,0);
              else
                rval = KeyControlPageDown(editor,1);
            break;
            case KEY_CPGUP:
              rval = KeyControlPageUp(editor,0);
            break;
            case KEY_SCPGUP:
              if (editor->viewonly)
                rval = KeyControlPageUp(editor,0);
              else
                rval = KeyControlPageUp(editor,1);
            break;
            case KEY_TOW:
            case KEY_CHOME:
              KeyTopOfWindow(editor);
            break;
            case KEY_BOW:
            case KEY_CEND:
              KeyBottomOfWindow(editor);
            break;
            case KEY_CTRLW:
            case KEY_CUP:
              rval = KeyScrollUp(editor);
            break;
            case KEY_CTRLZ:
            case KEY_CDN:
              rval = KeyScrollDown(editor);
            break;
            case KEY_CTRLY:
            case KEY_CDELETE:
              if (editor->viewonly == 0)
                rval = KeyDeleteLine(editor);
            break;
            case KEY_DELEOL:
            case KEY_ADELETE:
              if (editor->viewonly == 0)
                KeyDeleteEOL(editor);
            break;
            case KEY_CTRLT:
              if (editor->viewonly == 0)
                rval = KeyDeleteWord(editor);
            break;
            case KEY_CTRLG:
            case KEY_DELETE:
              if (editor->viewonly == 0)
                rval = KeyDelete(editor);
            break;
            case KEY_BLOCKDEL:
            case KEY_SDELETE:
              if (editor->viewonly == 0)
                BlockCut(editor);
            break;
            case KEY_TAB:
              if (editor->viewonly == 0)
                KeyTab(editor);
            break;
            case KEY_CTRLH:
            case KEY_BACKSPACE:
              if (editor->viewonly == 0)
                rval = KeyBackSpace(editor);
            break;
            case KEY_BEGINBLOCK:
              if (editor->viewonly == 0)
                BeginBlock(editor);
            break;
            case KEY_ENDBLOCK:
              if (editor->viewonly == 0)
                EndBlock(editor);
            break;
            case KEY_BLOCKLWR:
              if (editor->viewonly == 0)
                BlockUprLwr(editor,0);
            break;
            case KEY_BGNOFBLK:
              if (editor->viewonly == 0)
                BeginningOfBlk(editor);
            break;
            case KEY_ENDOFBLK:
              if (editor->viewonly == 0)
                EndOfBlk(editor);
            break;
            case KEY_BLOCKUPR:
              if (editor->viewonly == 0)
                BlockUprLwr(editor,1);
            break;
            case KEY_AINDENT:
              if (editor->viewonly == 0)
                AutoIndent(editor);
            break;
            case KEY_GOTOLINE:
              Cursor(2);
              rval = KeyGotoLine(editor);
              Cursor(0);
            break;
            case KEY_EXIT:
              if (editor->viewonly == 0)
                if (editor->macro_exe)
                  {
                    AbortMacro(editor);
                    editor->key = 0;
                  }
            break;
          }
        if (editor->viewonly == 0)
          {
            if (editor->macro_block)
              {
                if (editor->linenum < editor->p1line
                  || editor->linenum >=editor->p2line)
                  {
                    AbortMacro(editor);
                    MoveToLine(editor,editor->p1);
                    KeyHome(editor,0);
                  }
                else if (WBQueueIsEmpty(editor->keybuff))
                  ExecuteMacro(editor);
              }
            else if (editor->macro_exe)
              if (rval == 0 || WBQueueIsEmpty(editor->keybuff))
                AbortMacro(editor);
          }
        ResetOffsetPos(editor);
        Draw(editor);
      }
    while (editor->key != KEY_EXIT);
    EditCleanUp(editor);
  }

/*************************************************************************\
**  Main() simply processes the command line and then calls
**  VEditManager().
\*************************************************************************/
int main(int argc, char **argv)
  {
    WBERROR *errhdl;
    EDITOR *editor;
    WBQUEUE *keybuff;
    WBSTACK *backout;
    WBDLIST *cutbuffer;
    WBSLIST *macro;
    char charbuffer[MAXSTR+1], searchstr[MAXSRCHSTR+3];
    unsigned int srchoptions = 0;
    int i, viewonly = 0, tabspec = 8, entab = 0;

    if ((errhdl = WBErrorOpen(0,"error.txt")) != NULL)
      {
        WBMallocErrH(errhdl);
        if (argc == 1)
          {
            printf("\nUsage: vedt [-e] filename\n\n");
            printf("Where:\n");
            printf("       -e  = Entab when done editing\n");
            printf("       -tn = tab spec (tab size)\n");
            printf("             where n = the tab size\n");
            printf("       -v  = view mode (default is edit mode)\n\n");
            exit(1);
          }
        /*
        ** Get the current text attributes for hilighting.  Allowing
        ** the user to keep the same colors he/she is used to.
        */
        initscreen();

        for (i = 1; i < argc; i++)
          {
            if (argv[i][0] == '-')
              switch(argv[i][1])
                {
                  case 'e':              /* Entab file when done */
                  case 'E':
                    entab = 1;
                  break;
                  case 'v':              /* Don't allow editing */
                  case 'V':
                    viewonly = 1;
                  break;
                  case 't':              /* Set tab spec */
                  case 'T':
                    if (strlen(argv[i]+2))
                      tabspec = atoi(argv[i]+2);
                    else
                      tabspec = atoi(argv[++i]);
                  break;
                }
            else                         /* Must be a file for editing */
              {
                if ((keybuff = WBQueueOpen(NULL,sizeof(int),MAXKEYS,1)) != NULL)
                  {
                    if ((backout = WBStackOpen(NULL,sizeof(BACKOUT),MAXBACKOUT,1)) != NULL)
                      {
                        if ((macro = WBSListOpen(NULL)) != NULL)
                          {
                            if ((cutbuffer = WBDListOpen(NULL)) != NULL)
                              {
                                WBDListDeleteF(cutbuffer,NULL,(void (*)())DelFunct);
                                WBDListCopyF(cutbuffer,NULL,(void *(*)())CopyFunct);
                                if ((editor = EditorOpen(argv[i],viewonly,tabspec,entab)) != NULL)
                                  {
                                    memset(searchstr,0,MAXSRCHSTR+3);
                                    memset(charbuffer,0,MAXSTR+1);
                                    editor->keybuff = keybuff;
                                    editor->backoutstk = backout;
                                    editor->macro = macro;
                                    editor->dlistbuffer = cutbuffer;
                                    editor->searchstr = searchstr;
                                    editor->srchoptions = &srchoptions;
                                    editor->charbuffer = charbuffer;
                                    VEditManager(editor);
                                    EditorClose(editor);
                                  }
                                WBDListClose(cutbuffer);
                              }
                            WBSListClose(macro);
                          }
                        WBStackClose(backout);
                      }
                    WBQueueClose(keybuff);
                  }
                else printf("\nFailed to open the key buffer queue"), exit(1);
              }
          }
        WBErrorClose(errhdl);
      }
    return(0);
  }
