/*****************************************************************************\
**									     **
**  WW	   WW IIIIIIII NNN   NN	DDDDDDD	 BBBBBBB     AA	    SSSSSS EEEEEEEE  **
**  WW	W  WW	 II    NNNN  NN	DD    DD BB    BB  AA  AA  SS	   EE	     **
**  WW	W  WW	 II    NN NN NN	DD    DD BBBBBBB  AAAAAAAA  SSSSSS EEEEEE    **
**   WW	W WW	 II    NN  NNNN	DD    DD BB    BB AA	AA	SS EE	     **
**    WWWWW   IIIIIIII NN   NNN	DDDDDDD	 BBBBBBB  AA	AA SSSSSS  EEEEEEEE  **
**									     **
**   SSSSSS  OOOOOO  FFFFFFFF TTTTTTTT WW     WW    AA	  RRRRRRR  EEEEEEEE  **
**  SS	    OO	  OO FF		 TT    WW  W  WW  AA  AA  RR	RR EE	     **
**   SSSSS  OO	  OO FFFFF	 TT    WW  W  WW AAAAAAAA RRRRRRR  EEEEEE    **
**	 SS OO	  OO FF		 TT	WW W WW	 AA    AA RR   RR  EE	     **
**  SSSSSS   OOOOOO  FF		 TT	 WWWWW	 AA    AA RR	RR EEEEEEEE  **
**									     **
*********** NOTICE ************************************************************
**	  This file contains valuable trade secrets and	proprietary	     **
**	  assets of Windbase Software Inc.  Embodying substantial	     **
**	  creative efforts and confidential information.  Unauthorized	     **
**	  use, copying,	decompiling, translating, disclosure or		     **
**	  transfer, of any kind, is strictly prohibited.		     **
**									     **
**	  COPYRIGHT (C)	1992, 1993, 1994.  Windbase Software Inc.	     **
**	  ALL RIGHTS RESERVED.						     **
\*****************************************************************************/


/*************************************************************************\
**  Global defines.
\*************************************************************************/

#define	MAXBACKOUT 256
#define	MAXKEYS	256
#define	MAXSTR 256
#define	MAXSRCHSTR 60
#define	MAXSCRROW 25
#define	MAXSCRCOL 80
#define	MAXROW (MAXSCRROW-2)
#define	MAXCOL (MAXSCRCOL-1)
#define	PTR1	 0x01
#define	PTR2	 0x02
#define	BOTHPTRS 0x03
#define	NOPTR1	 0x04
#define	NOPTR2	 0x08
#define	NOPTRS	 0x0C

#define	SRCH_FORWARD 0x01
#define	SRCH_IGNORECASE	0x02
#define	SRCH_PROMPT 0x04
#define	SRCH_REPLACEALL	0x08


/*************************************************************************\
**  Backout struct for incremental undo
\*************************************************************************/

#define	BK_INSERT 0x0001

typedef	struct backout_buf
  {
    int	flags,
	key,
	overkey,
	pos,
	offset;
    long linenum;
    char *buf;
  } BACKOUT;


/*************************************************************************\
**  Main editor	structure.
\*************************************************************************/

typedef	struct editor_st
  {
    BACKOUT backout;
    WBQUEUE *keybuff;
    WBSTACK *backoutstk;
    WBSLIST *macro,
	    *filelist;
    WBDLIST *dlistbuffer,
	    *dlist,
	    *dltopscr,
	    *p1,
	    *p2;
    char *charbuffer,
	 tabstr[MAXSTR+1],
	 delims[MAXSTR+1],
	 filename[MAXSTR+1],
	 str[MAXSTR+1],
	 *strptr,
	 *searchstr;

    long linenum,
	 p1line,
	 p2line;

    unsigned int *srchoptions;
    int	tabspec,
	indent,
	entab,
	offset,
	row,
	pos,
	p1pos,
	p2pos,
	insert,
	changed,
	chgmade,
	realloc,
	key,
	ch,
	viewonly,
	deltop,
	instop,
	refillscr,
	fillstart,
	fillend,
	hideblock,
	macro_on,
	macro_exe,
	macro_block,
	help;
  } EDITOR;


/*************************************************************************\
**  Editor function prototypes.
\*************************************************************************/

char *insstr(char *str1, char *str2, int pos);
char *UpdateCurrent(EDITOR *editor);
char *UpdateLine(WBDLIST *dlist, char *str, int	realloc);
EDITOR *EditorOpen(char	*filename, int viewonly, int tabspec, int entab);
int AddToMacro(EDITOR *editor, int keycombo);
int KeyBackSpace(EDITOR	*editor);
int KeyBackSpaceFirst(EDITOR *editor);
int KeyControlLeft(EDITOR *editor, int shift);
int KeyControlPageDown(EDITOR *editor, int shift);
int KeyControlPageUp(EDITOR *editor, int shift);
int KeyControlRight(EDITOR *editor, int	shift);
int KeyDelete(EDITOR *editor);
int KeyDeleteEnd(EDITOR	*editor);
int KeyDeleteLine(EDITOR *editor);
int KeyDeleteWord(EDITOR *editor);
int KeyDown(EDITOR *editor, int	shift);
int KeyLeft(EDITOR *editor, int	shift);
int KeyPageDown(EDITOR *editor,	int shift);
int KeyPageUp(EDITOR *editor, int shift);
int KeyRight(EDITOR *editor, int shift);
int KeyScrollDown(EDITOR *editor);
int KeyScrollUp(EDITOR *editor);
int KeyUp(EDITOR *editor, int shift);
int main(int argc, char	**argv);
int Search(EDITOR *editor, int searchfirst);
int SearchBackward(EDITOR *editor);
int SearchForward(EDITOR *editor);
unsigned int ChkBlkPtrs(EDITOR *editor);
void *CopyFunct(void *nullitem,	char *item);
void AbortMacro(EDITOR *editor);
void AbortMacro(EDITOR *editor);
void BeginBlock(EDITOR *editor);
void BlockCopy(EDITOR *editor);
void BlockCut(EDITOR *editor);
void BlockInsert(EDITOR	*editor);
void BlockUprLwr(EDITOR	*editor, int upper);
void CloseBlock(EDITOR *editor);
void Cursor(int	cursorsize);
void Detab(char	*str1, char *str2, int maxlen, char *tabstr);
void Entab(char	*str1, char *str2, int maxlen, char *tabstr);
void Draw(EDITOR *editor);
void EditCleanUp(EDITOR	*editor);
void EditorClose(EDITOR	*editor);
void EndBlock(EDITOR *editor);
void ExecuteMacro(EDITOR *editor);
void ExecuteMacroBlock(EDITOR *editor);
void FillScreen(EDITOR *editor);
void DelFunct(void *nullitem, void *item);
void HideBlock(EDITOR *editor);
void InitEditor(EDITOR *editor);
void InitTabs(char *tabstr, int	maxlen,	int tabspec);
void KeyBottomOfWindow(EDITOR *editor);
void KeyDeleteEOL(EDITOR *editor);
void KeyEnd(EDITOR *editor, int	shift);
void KeyEnter(EDITOR *editor);
void KeyHome(EDITOR *editor, int shift);
void KeyInsert(EDITOR *editor);
void KeyInsertLine(EDITOR *editor);
void KeyTab(EDITOR *editor);
void KeyTopOfWindow(EDITOR *editor);
void LineMove(EDITOR *editor);
void LoadFile(EDITOR *editor);
void Macro(EDITOR *editor);
void MoveToLine(EDITOR *editor,	WBDLIST	*lineptr);
void ProcessChar(EDITOR	*editor);
void ProcessKey(EDITOR *editor);
void PutFile(EDITOR *editor, WBDLIST *dlist, char *filename, int backup);
void ReadBlock(EDITOR *editor, char *filename);
void ReadMacro(EDITOR *editor);
void ResetOffsetPos(EDITOR *editor);
void SaveFile(EDITOR *editor);
void SaveFileAs(EDITOR *editor);
void SetBlkPtrs(EDITOR *editor,	unsigned int ptrbits);
void SetBlock(EDITOR *editor);
void UpdateCoordinates(EDITOR *editor, int start, int end);
void UpdateStats(EDITOR	*editor, int updateall);
void VEditManager(EDITOR *editor);
void WriteBlock(EDITOR *editor);
void WriteMacro(EDITOR *editor);
WBDLIST	*GetFile(EDITOR	*editor, char *filename);
int BeginningOfBlk(EDITOR *editor);
int EndOfBlk(EDITOR *editor);
void AutoIndent(EDITOR *editor);
void SearchBack(EDITOR *editor);
int SearchReplace(EDITOR *editor);
char *wbstrstr(char *str1, char	*str2);
char *wbstrstri(char *str1, char *str2);
int wbstrcmp(char *str1, char *str2);
int wbstrcmpi(char *str1, char *str2);
int wbstrncmp(char *str1, char *str2, int len);
int wbstrncmpi(char *str1, char	*str2, int len);
char *srchstrstr(char *str1, char *str2, int ignorecase);
int srchstrcmp(char *str1, char	*str2, int ignorecase);
int srchstrncmp(char *str1, char *str2,	int len, int ignorecase);
int KeyGotoLine(EDITOR *editor);
void AddBackout(EDITOR *editor);
